// Camera.cpp
//		Camera class intended for use with OpenGL ES 2.0
//		Compiled 2/10/2011 by Matt Pierce
#include <stdlib.h>
#include <string.h>
#include "VectorMath.h"
#include "Camera.h"

// function UpdateVectors
//		arguments: 	none
//		returns: 	void
// Updates internal data used to generate the transformation matrix
void CCamera::UpdateVectors() {
	CMatrix RollMatrix;
	if ((pos != lookat) && (up != v3d(0,0,0))){
		// this may seem odd at first, but this is intened to generate negative values
		vfwd = vmNormalize(pos - lookat);
		vright = vmNormalize(up)^vfwd;
		vup = vfwd^vright;
		
		RollMatrix.LoadIdentity();
		RollMatrix.Rotate(-roll, vfwd.x, vfwd.y, vfwd.z);
		// vfwd = RollMatrix.Transform(vfwd);
		vright = RollMatrix.Transform(vright);
		vup = RollMatrix.Transform(vup);
	}
}

// function SetPos
//		arguments:	const SVect3d &aPos			the position at which the camera should be placed
//		returns:	void
// Sets the camera's position
void CCamera::SetPos(const SVect3d &aPos) {
	pos = aPos;
	UpdateVectors();
}

// function SetLookAt
//		arguments:	const SVect3d &aLookAt		the coordinate on which the camera is focused
//		returns:	void
// Sets the position at which the camera is looking  (not the camera's direction)
void CCamera::SetLookAt(const SVect3d &aLookAt) {
	lookat = aLookAt;
	UpdateVectors();
}

// function SetUp
//		arguments:	const SVect3d &aUp			the direction that will appear as "up" on the screen
//		returns: void
// Sets the direction that will appear ss "up" on the screen.  This exact direction is not always guaranteed
// to be "up," but rather the calculated "up" will be in the plane that contains both up and (lookat-pos).
void CCamera::SetUp(const SVect3d &aUp) {
	up = aUp;
	UpdateVectors();
}

// function SetRoll
//		arguments: const float &aRoll			an angle in radians
//		returns: void
// Sets a roll value independent of the Up vector.  This is used to allow for easy screen orientation adjustment
// without having to completely re-tool your camera 
void CCamera::SetRoll(const float &aRoll) {
	roll = aRoll;
	UpdateVectors();
}

// function GetMatrixArray
//		arguments:	float *tMat					this variable gets set with the transformation matrix values and must be allocated ahead of time by the user
//		returns:	void
// Uses the calculated values to create an OpenGL ES 2.0 compatible matrix that the user can put in a vertex shader
void CCamera::GetMatrixArray(float *tMat) {
	tMat[0]  = vright.x;
	tMat[1]  = vup.x;
	tMat[2]  = vfwd.x;
	tMat[3]  = 0;
	
	tMat[4]  = vright.y;
	tMat[5]  = vup.y;
	tMat[6]  = vfwd.y;
	tMat[7]  = 0;
	
	tMat[8]  = vright.z;
	tMat[9]  = vup.z;
	tMat[10] = vfwd.z;
	tMat[11] = 0;
	
	tMat[12] = -(vright*pos);
	tMat[13] = -(vup*pos);
	tMat[14] = -(vfwd*pos);
	tMat[15] = 1;
	
	
}

// Standard GL perspective matrix creation
void Persp(float Proj[4][4], const float w, const float h, const float FOV, const float ZNear, const float ZFar)
{
    const float Delta   = ZFar - ZNear;

    memset(Proj, 0, 4*4*sizeof(float));

    Proj[0][0] = 1.0f / tanf(FOV * 3.1415926535f / 360.0f);
    Proj[1][1] = w*Proj[0][0]/h;

    Proj[2][2] = -(ZFar + ZNear) / Delta;
    Proj[2][3] = -1.0f;
    Proj[3][2] = -2.0f * ZFar * ZNear / Delta;
}

// Standard GL orthographic matrix creation
void Ortho(float Ortho[4][4], const float left, const float right, const float bottom, const float top, const float nearVal, const float farVal)
{
	if ((right != left) && (top != bottom) && (nearVal != farVal)) {
		memset(Ortho, 0, 4*4*sizeof(float));
		
		Ortho[0][0] = 2.0f/(right-left);
		Ortho[1][1] = 2.0f/(top-bottom);
		Ortho[2][2] = 2.0f/(farVal-nearVal);
		Ortho[3][3] = 1.0f;
		
		// Row Major
		Ortho[0][3] = - (right+left)/(right-left);
		Ortho[1][3] = - (top+bottom)/(top-bottom);
		Ortho[2][3] = - (farVal+nearVal)/(farVal-nearVal);
	}
}
