// VectorMath.cpp
//		Platform-independent library of 3D vector and matrix tools
//		Compiled 2/10/2011 by Matt Pierce
#include <math.h>
#include <stdlib.h>
#include "VectorMath.h"

const float pi = 3.1415926535897932384626433832795;

SVect3d v3d(float x, float y, float z)
{
	SVect3d t;
	t.x = x; t.y = y; t.z = z;
	return t;
}

float vmDistance(const SVect3d &a, const SVect3d &b)
{
	SVect3d t;
	t.x = b.x - a.x;
	t.y = b.y - a.y;
	t.z = b.z - a.z;
	return sqrtf((float)(t.x*t.x + t.y*t.y + t.z*t.z));
}

float vmLength(const SVect3d &a)
{
	return sqrtf((float)(a.x*a.x + a.y*a.y + a.z*a.z));
}

SVect3d vmNormalize(const SVect3d &v)
{
	float d = vmDistance(v, v3d(0,0,0));
	SVect3d t;
	t.x = v.x/d;
	t.y = v.y/d;
	t.z = v.z/d;
	return t;
}

SVect3d vmCrossProduct(const SVect3d &a, const SVect3d &b)
{
	SVect3d t;
	t.x = a.y*b.z - a.z*b.y;
	t.y = a.z*b.x - a.x*b.z;
	t.z = a.x*b.y - a.y*b.x;
	return t;
}

float vmDotProduct(const SVect3d &a, const SVect3d &b)
{
	return a.x*b.x + a.y*b.y + a.z*b.z;
}

float vmGetAngle(const SVect3d &a, const SVect3d &b)
{
	if (a == b) return 0;
	float dp = vmDotProduct(a,b);
	if (dp>0.99999999) return 0;
	else if (dp<-0.99999999) return pi;
	else return acos(dp);
}

SQuaternion4d vmConjugate(const SQuaternion4d &q)
{
	SQuaternion4d q2;
	q2.x = -q.x;
	q2.y = -q.y;
	q2.z = -q.z;
	q2.w = q.w;
	return q2;
}

SQuaternion4d vmNormalize(const SQuaternion4d &q)
{
	SQuaternion4d q2;
	float n = sqrtf((float)(q.x*q.x + q.y*q.y + q.z*q.z + q.w*q.w));
	q2.x = q.x/n;
	q2.y = q.y/n;
	q2.z = q.z/n;
	q2.w = q.w/n;
	return q2;
}

SQuaternion4d vmInterpolate(const SQuaternion4d &a, const SQuaternion4d &b, const float t)
{
	// SQuaternion4d q;
	// float HalfTheta = acos(a.w*b.w + a.x*b.x + a.y*b.y + a.z*b.z);
	// q = (a*sin((1-t)*HalfTheta) + b*sin(t*HalfTheta))*(1.0/sin(HalfTheta));
	// return q;
	return a;
}

const SVect3d operator-(const SVect3d &v)
{
	SVect3d t;
	t.x = -v.x;
	t.y = -v.y;
	t.z = -v.z;
	return t;
}

const SVect3d operator-(const SVect3d &a, const SVect3d &b)
{
	SVect3d t;
	t.x = a.x - b.x;
	t.y = a.y - b.y;
	t.z = a.z - b.z;
	return t;
}

const SVect3d operator+(const SVect3d &a, const SVect3d &b)
{
	SVect3d t;
	t.x = a.x + b.x;
	t.y = a.y + b.y;
	t.z = a.z + b.z;
	return t;
}

const SVect3d operator*(const SVect3d &a, const float &b)
{
	SVect3d t;
	t.x = a.x*b;
	t.y = a.y*b;
	t.z = a.z*b;
	return t;
}

const SVect3d operator/(const SVect3d &a, const float &b)
{
	SVect3d t;
	t.x = a.x/b;
	t.y = a.y/b;
	t.z = a.z/b;
	return t;
}

SQuaternion4d operator*(const SQuaternion4d &a, const SQuaternion4d &b)
{
	SQuaternion4d q;
	q.x = a.x*b.w + a.y*b.z - a.z*b.y + a.w*b.x;
	q.y =-a.x*b.z + a.y*b.w + a.z*b.x + a.w*b.y;
	q.z = a.x*b.y - a.y*b.x + a.z*b.w + a.w*b.z;
	q.w =-a.x*b.x - a.y*b.y - a.z*b.z + a.w*b.w;
	return q;
}

SQuaternion4d operator*(const SQuaternion4d &a, const float &b)
{
	SQuaternion4d q;
	q.x = a.x*b;
	q.y = a.y*b;
	q.z = a.z*b;
	q.w = a.w*b;
	return q;
}
SQuaternion4d operator+(const SQuaternion4d &a, const SQuaternion4d &b)
{
	SQuaternion4d q;
	q.x = a.x + b.x;
	q.y = a.y + b.y;
	q.z = a.z + b.z;
	q.w = a.w + b.w;
	return q;
}

SVect3d vmReflect(const SVect3d &view, const SVect3d &normal)
{
	return (-2.0 * vmDotProduct(view, normal))*normal + view;
}

void CMatrix::LoadIdentity()
{
	for(int i=0; i<16; i++)
	{
		lpfMatrix[i] = lpfInverse[i] = 0.0;
	}

	lpfMatrix[0] = lpfInverse[0] = 1.0;
	lpfMatrix[5] = lpfInverse[5] = 1.0;
	lpfMatrix[10]= lpfInverse[10]= 1.0;
	lpfMatrix[15]= lpfInverse[15]= 1.0;

	bInverseUpToDate = true;
}

void CMatrix::Push()
{
	SMatrixStack *temp = lpsmsNext;
	lpsmsNext = new SMatrixStack;
	for (int i=0;i<16;i++)
	{
		lpsmsNext->lpfMatrix[i] = lpfMatrix[i];
		lpsmsNext->lpfInverse[i] = lpfInverse[i];
	}
	lpsmsNext->bInverseUpToDate = bInverseUpToDate;
	lpsmsNext->lpsmsNext = temp;
}

void CMatrix::Pop()
{
	SMatrixStack *temp = lpsmsNext;
	if (lpsmsNext == NULL)
	{
		LoadIdentity();
	} else {
		for (int i=0;i<16;i++)
		{
			lpfMatrix[i] = lpsmsNext->lpfMatrix[i];
			lpfInverse[i] = lpsmsNext->lpfInverse[i];
		}
		bInverseUpToDate = lpsmsNext->bInverseUpToDate;
		lpsmsNext = lpsmsNext->lpsmsNext;
		delete temp;
	}

}

void CMatrix::Concatenate(float *m)
{
	float tm[16];
	int j;
	
	for(j=0; j<16; j++) tm[j] = 0.0f;
	tm[0] = tm[5] = tm[10] = tm[15] = 1.0f;

	for (int i=0; i<4; i++)
	{
		j = 4*i;
		tm[j+0] = m[j]*lpfMatrix[0] + m[j+1]*lpfMatrix[4] + m[j+2]*lpfMatrix[8] + m[j+3]*lpfMatrix[12];
		tm[j+1] = m[j]*lpfMatrix[1] + m[j+1]*lpfMatrix[5] + m[j+2]*lpfMatrix[9] + m[j+3]*lpfMatrix[13];
		tm[j+2] = m[j]*lpfMatrix[2] + m[j+1]*lpfMatrix[6] + m[j+2]*lpfMatrix[10]+ m[j+3]*lpfMatrix[14];
		tm[j+3] = m[j]*lpfMatrix[3] + m[j+1]*lpfMatrix[7] + m[j+2]*lpfMatrix[11]+ m[j+3]*lpfMatrix[15];
	}
	for (j=0;j<16;j++)  lpfMatrix[j] = tm[j];
	bInverseUpToDate = false;
}

void CMatrix::Rotate(float angle, float x, float y, float z)
{
	float tm[16];
	int j;
	for(j=0; j<16; j++) tm[j] = 0.0f;
	tm[0] = tm[5] = tm[10] = tm[15] = 1.0f;

	float c = cos(angle);
	float s = sin(angle);
	float v = 1.0f - c;

	/*tm[0] = 1.0f+(1.0f-c)*(x*x-1.0f);
	tm[1] = -z*s + (1.0f-c)*x*y;
	tm[2] = y*s + (1.0f-c)*x*z;
	tm[4] = z*s + (1.0f-c)*x*y;
	tm[5] = 1.0f+(1.0f-c)*(y*y-1.0f);
	tm[6] = -x*s + (1.0f-c)*y*z;
	tm[8] = -y*s + (1.0f-c)*x*z;
	tm[9] = x*s + (1.0f-c)*y*z;
	tm[10]= 1.0f + (1.0f-c)*(z*z-1.0f);*/

	tm[0] = x*x*v + c;
	tm[1] = x*y*v - z*s;
	tm[2] = x*z*v + y*s;
	// tm[4] = x*y + z*s; // old code
	tm[4] = x*y*v + z*s;
	tm[5] = y*y*v + c;
	tm[6] = y*z*v - x*s;
	tm[8] = x*z*v - y*s;
	tm[9] = y*z*v + x*s;
	tm[10]= z*z*v + c;

	Concatenate(tm);
}

void CMatrix::Scale(float x, float y, float z)
{
	float tm[16];
	int j;
	for(j=0; j<16; j++) tm[j] = 0.0;
	tm[0] = tm[5] = tm[10] = tm[15] = 1.0;

	tm[0] = x;
	tm[5] = y;
	tm[10]= z;
	Concatenate(tm);
}

void CMatrix::Translate(float x, float y, float z)
{
	float tm[16];
	int j;
	for(j=0; j<16; j++) tm[j] = 0.0;
	tm[0] = tm[5] = tm[10] = tm[15] = 1.0;

	tm[3] = x;
	tm[7] = y;
	tm[11]= z;
	
	Concatenate(tm);
}

SVect3d CMatrix::Transform(const SVect3d &v)
{
	SVect3d t;

	t.x = v.x*lpfMatrix[0] + v.y*lpfMatrix[1] + v.z*lpfMatrix[2] + lpfMatrix[3];
	t.y = v.x*lpfMatrix[4] + v.y*lpfMatrix[5] + v.z*lpfMatrix[6] + lpfMatrix[7];
	t.z = v.x*lpfMatrix[8] + v.y*lpfMatrix[9] + v.z*lpfMatrix[10]+ lpfMatrix[11];

	return t;
}

SVect3d CMatrix::InverseTransform(const SVect3d &v)
{
	SVect3d t;

	if (!bInverseUpToDate) UpdateInverse();

	t.x = v.x*lpfInverse[0] + v.y*lpfInverse[1] + v.z*lpfInverse[2] + lpfInverse[3];
	t.y = v.x*lpfInverse[4] + v.y*lpfInverse[5] + v.z*lpfInverse[6] + lpfInverse[7];
	t.z = v.x*lpfInverse[8] + v.y*lpfInverse[9] + v.z*lpfInverse[10]+ lpfInverse[11];

	return t;
}

SVect3d CMatrix::NormalTransform(const SVect3d &v)
{
	SVect3d t;

	if (!bNormalUpToDate) UpdateNormal();

	t.x = v.x*lpfNormal[0] + v.y*lpfNormal[1] + v.z*lpfNormal[2] + lpfNormal[3];
	t.y = v.x*lpfNormal[4] + v.y*lpfNormal[5] + v.z*lpfNormal[6] + lpfNormal[7];
	t.z = v.x*lpfNormal[8] + v.y*lpfNormal[9] + v.z*lpfNormal[10]+ lpfNormal[11];

	return t;
}

SVect3d CMatrix::InverseNormalTransform(const SVect3d &v)
{
	SVect3d t;

	if (!bInverseNormalUpToDate) UpdateInverseNormal();

	t.x = v.x*lpfInverseNormal[0] + v.y*lpfInverseNormal[1] + v.z*lpfInverseNormal[2] + lpfInverseNormal[3];
	t.y = v.x*lpfInverseNormal[4] + v.y*lpfInverseNormal[5] + v.z*lpfInverseNormal[6] + lpfInverseNormal[7];
	t.z = v.x*lpfInverseNormal[8] + v.y*lpfInverseNormal[9] + v.z*lpfInverseNormal[10]+ lpfInverseNormal[11];

	return t;
}

void CMatrix::UpdateInverse()
{
	Invert(lpfMatrix, lpfInverse);
	bInverseUpToDate = true;
}

void CMatrix::UpdateNormal()
{
	float temp[16];
	// Transpose(lpfMatrix, temp);
	// Invert(temp, lpfNormal);
	Invert(lpfMatrix, temp);
	Transpose(temp, lpfNormal);
	// Transpose(lpfMatrix, lpfNormal);
	bNormalUpToDate = true;
}

void CMatrix::UpdateInverseNormal()
{
	if (!bNormalUpToDate) UpdateNormal();
//	float temp[16];
//	Transpose(lpfInverse, temp);
//	Invert(temp, lpfInverseNormal);
	Invert(lpfNormal, lpfInverseNormal);
	bInverseNormalUpToDate = true;
}

void CMatrix::Transpose(float *mat, float *tmat)
{
	int x, y;
	for(y=0;y<4;y++) {
		for (x=0;x<4;x++) {
			tmat[x*4 + y] = mat[y*4 + x];
		}
	}
}

void CMatrix::Invert(float *mat, float *imat)
{
	float tmp[12];
	float src[16];
	float det;

	for (int i=0; i<4; i++)
	{
		src[i] = mat[i*4];
		src[i+4] = mat[i*4+1];
		src[i+8] = mat[i*4+2];
		src[i+12]= mat[i*4+3];
	}

	tmp[0] = src[10]*src[15];
	tmp[1] = src[11]*src[14];
	tmp[2] = src[9]*src[15];
	tmp[3] = src[11]*src[13];
	tmp[4] = src[9]*src[14];
	tmp[5] = src[10]*src[13];
	tmp[6] = src[8]*src[15];
	tmp[7] = src[11]*src[12];
	tmp[8] = src[8]*src[14];
	tmp[9] = src[10]*src[12];
	tmp[10] = src[8]*src[13];
	tmp[11] = src[9]*src[12];

	/*calculate first 8 elements (cofactors)*/
	imat[0] =  tmp[0]*src[5] + tmp[3]*src[6] + tmp[4]*src[7];
	imat[0] -= tmp[1]*src[5] + tmp[2]*src[6] + tmp[5]*src[7];
	imat[1] =  tmp[1]*src[4] + tmp[6]*src[6] + tmp[9]*src[7];
	imat[1] -= tmp[0]*src[4] + tmp[7]*src[6] + tmp[8]*src[7];
	imat[2] =  tmp[2]*src[4] + tmp[7]*src[5] + tmp[10]*src[7];
	imat[2] -= tmp[3]*src[4] + tmp[6]*src[5] + tmp[11]*src[7];
	imat[3] =  tmp[5]*src[4] + tmp[8]*src[5] + tmp[11]*src[6];
	imat[3] -= tmp[4]*src[4] + tmp[9]*src[5] + tmp[10]*src[6];
	imat[4] =  tmp[1]*src[1] + tmp[2]*src[2] + tmp[5]*src[3];
	imat[4] -= tmp[0]*src[1] + tmp[3]*src[2] + tmp[4]*src[3];
	imat[5] =  tmp[0]*src[0] + tmp[7]*src[2] + tmp[8]*src[3];
	imat[5] -= tmp[1]*src[0] + tmp[6]*src[2] + tmp[9]*src[3];
	imat[6] =  tmp[3]*src[0] + tmp[6]*src[1] + tmp[11]*src[3];
	imat[6] -= tmp[2]*src[0] + tmp[7]*src[1] + tmp[10]*src[3];
	imat[7] =  tmp[4]*src[0] + tmp[9]*src[1] + tmp[10]*src[2];
	imat[7] -= tmp[5]*src[0] + tmp[8]*src[1] + tmp[11]*src[2];

	/*calculate pairs for second 8 elements (cofactors)*/
	tmp[0] = src[2]*src[7];
	tmp[1] = src[3]*src[6];
	tmp[2] = src[1]*src[7];
	tmp[3] = src[3]*src[5];
	tmp[4] = src[1]*src[6];
	tmp[5] = src[2]*src[5];
	tmp[6] = src[0]*src[7];
	tmp[7] = src[3]*src[4];
	tmp[8] = src[0]*src[6];
	tmp[9] = src[2]*src[4];
	tmp[10]= src[0]*src[5];
	tmp[11]= src[1]*src[4];

	/*calculate second 8 elements (cofactors)*/
	imat[8] =  tmp[0]*src[13] + tmp[3]*src[14] + tmp[4]*src[15];
	imat[8] -= tmp[1]*src[13] + tmp[2]*src[14] + tmp[5]*src[15];
	imat[9] =  tmp[1]*src[12] + tmp[6]*src[14] + tmp[9]*src[15];
	imat[9] -= tmp[0]*src[12] + tmp[7]*src[14] + tmp[8]*src[15];
	imat[10]=  tmp[2]*src[12] + tmp[7]*src[13] + tmp[10]*src[15];
	imat[10]-= tmp[3]*src[12] + tmp[6]*src[13] + tmp[11]*src[15];
	imat[11]=  tmp[5]*src[12] + tmp[8]*src[13] + tmp[11]*src[14];
	imat[11]-= tmp[4]*src[12] + tmp[9]*src[13] + tmp[10]*src[14];
	imat[12]=  tmp[2]*src[10] + tmp[5]*src[11] + tmp[1]*src[9];
	imat[12]-= tmp[4]*src[11] + tmp[0]*src[9] + tmp[3]*src[10];
	imat[13]=  tmp[8]*src[11] + tmp[0]*src[8] + tmp[7]*src[10];
	imat[13]-= tmp[6]*src[10] + tmp[9]*src[11] + tmp[1]*src[8];
	imat[14]=  tmp[6]*src[9] + tmp[11]*src[11] + tmp[3]*src[8];
	imat[14]-= tmp[10]*src[11] + tmp[2]*src[8] + tmp[7]*src[9];
	imat[15]=  tmp[10]*src[10] + tmp[4]*src[8] + tmp[9]*src[9];
	imat[15]-= tmp[8]*src[9] + tmp[11]*src[10] + tmp[5]*src[8];

	/*calculate determinant*/
	det = src[0]*lpfInverse[0] + src[1]*lpfInverse[1] + src[2]*lpfInverse[2] + src[3]*lpfInverse[3];

	/*calculate matrix inverse*/
	if (det != 0)	det = 1.0/det;
	else det = 1.0;
	for (int j=0;j<16;j++)
	{
		imat[j] *= det;
	}
}

void CMatrix::Duplicate(const CMatrix &m)
{
	bInverseUpToDate = false;
	int i;
	for (i=0;i<16;i++) {
		lpfMatrix[i] = m.lpfMatrix[i];
	}
}

void CMatrix::FromQuaternion(const SQuaternion4d &q)
{
	LoadIdentity();
	// first row
	lpfMatrix[0] = q.x*q.x + q.w*q.w - q.y*q.y - q.z*q.z;
	lpfMatrix[1] = 2*q.x*q.y - 2*q.w*q.z;
	lpfMatrix[2] = 2*q.x*q.z + 2*q.w*q.w;
	
	// second row
	lpfMatrix[4] = 2*q.w*q.z + 2*q.x*q.y;
	lpfMatrix[5] = q.w*q.w -q.x*q.x + q.y*q.y -q.z*q.z;
	lpfMatrix[6] = -2*q.w*q.x + 2*q.y*q.z;

	// third row
	lpfMatrix[8] = -2*q.w*q.y + 2*q.x*q.z;
	lpfMatrix[9] = 2*q.w*q.x + 2*q.y*q.z;
	lpfMatrix[10] = q.w*q.w - q.x*q.x - q.y*q.y + q.z*q.z;
}

void CMatrix::GetMatrixArray(float *tMat)
{
	Transpose(lpfMatrix, tMat);
}

bool vmSameSide(const SVect3d &p, const SVect3d &a, const SVect3d &b, const SVect3d &n)
{
	SVect3d cp = (b-a)^(p-a);
	return (cp*n >= 0);
}

bool vmInTriangle(const SVect3d &p, const SVect3d &v1, const SVect3d &v2, const SVect3d &v3)
{
	SVect3d n = (v2-v1)^(v3-v1);
	return (vmSameSide(p, v1, v2, n)&&vmSameSide(p, v2, v3, n)&&vmSameSide(p, v3, v1, n));
}
