// VectorMath.h
//		Platform-independent library of 3D vector and matrix tools
//		Compiled 2/10/2011 by Matt Pierce
#pragma once
#include <stdlib.h>
#include <math.h>

// Vector Stuff

struct SVect3d {float x, y, z;};
struct SQuaternion4d {float x, y, z, w;};

#define expandv3d(v) (v).x, (v).y, (v).z
#define macrodot(a, b) a.x*b.x + a.y*b.y + a.z*b.z

// basic 3d math
SVect3d v3d(float x, float y, float z);
SVect3d vmNormalize(const SVect3d &v);
SVect3d vmCrossProduct(const SVect3d &a, const SVect3d &b);
SVect3d vmReflect(const SVect3d &view, const SVect3d &normal);
float vmDistance(const SVect3d &a, const SVect3d &b);
float vmLength(const SVect3d &a);
float vmDotProduct(const SVect3d &a, const SVect3d &b);
float vmGetAngle(const SVect3d &a, const SVect3d &b);
SQuaternion4d vmConjugate(const SQuaternion4d &q);
SQuaternion4d vmNormalize(const SQuaternion4d &q);
SQuaternion4d vmInterpolate(const SQuaternion4d &a, const SQuaternion4d &b, const float t);

// helpful algorithms
bool vmSameSide(const SVect3d &p, const SVect3d &a, const SVect3d &b, const SVect3d &n);
bool vmInTriangle(const SVect3d &p, const SVect3d &v1, const SVect3d &v2, const SVect3d &v3);

// unary negation
const SVect3d operator-(const SVect3d &v);
// vector - vector
const SVect3d operator-(const SVect3d &a, const SVect3d &b);
// vector + vector
const SVect3d operator+(const SVect3d &a, const SVect3d &b);
// vector * scalar
const SVect3d operator*(const SVect3d &a, const float &b);
inline const SVect3d operator*(const float &a, const SVect3d &b)
{ return b*a;}
// vector / scalar
const SVect3d operator/(const SVect3d &a, const float &b);
// dot product shortcut
inline float operator*(const SVect3d &a, const SVect3d &b)
{ return vmDotProduct(a, b); }
// cross product shortcut
inline const SVect3d operator^(const SVect3d &a, const SVect3d &b)
{ return vmCrossProduct(a, b);}
inline const bool operator==(const SVect3d &a, const SVect3d &b)
{ return (a.x==b.x)&&(a.y==b.y)&&(a.z==b.z);}
inline const bool operator!=(const SVect3d &a, const SVect3d &b)
{ return !(a==b);}
// Quaternion operators
SQuaternion4d operator*(const SQuaternion4d &a, const SQuaternion4d &b);
SQuaternion4d operator*(const SQuaternion4d &a, const float &b);
inline SQuaternion4d operator*(const float &a, const SQuaternion4d &b)
{return b*a;}
SQuaternion4d operator+(const SQuaternion4d &a, const SQuaternion4d &b);




// Matrix Stuff

struct SMatrixStack {
	float lpfMatrix[16];
	float lpfInverse[16];
	bool bInverseUpToDate;
	SMatrixStack *lpsmsNext;
};

class CMatrix {
public:
	CMatrix():lpsmsNext(NULL), bInverseUpToDate(false), bNormalUpToDate(false), bInverseNormalUpToDate(false) {LoadIdentity();};
	~CMatrix() {SMatrixStack *tmp; while (lpsmsNext) {tmp = lpsmsNext; lpsmsNext = lpsmsNext->lpsmsNext; delete tmp;}};

	void LoadIdentity();
	void Rotate(float angle, float x, float y, float z);
	void Translate(float x, float y, float z);
	void Scale(float x, float y, float z);
	void Push();
	void Pop();
	void Duplicate(const CMatrix &m);
	void FromQuaternion(const SQuaternion4d &q);

	SVect3d Transform(const SVect3d &v);
	SVect3d InverseTransform(const SVect3d &v);
	SVect3d NormalTransform(const SVect3d &v);
	SVect3d InverseNormalTransform(const SVect3d &v);
	
	void GetMatrixArray(float *tMat);

private:
	float lpfMatrix[16];
	float lpfInverse[16];
	float lpfNormal[16];
	float lpfInverseNormal[16];
	bool bInverseUpToDate;
	bool bNormalUpToDate;
	bool bInverseNormalUpToDate;
	SMatrixStack *lpsmsNext;
	void Invert(float *mat, float *imat);
	void Transpose(float *mat, float *tmat);
	void UpdateInverse();
	void UpdateNormal();
	void UpdateInverseNormal();
	void Concatenate(float *m);
};
